/*
 * Copyright (c) 2017, Norman Gray
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * pqrst 1.0, revision 35be58e470c3, of 2018 August 18
 * See https://bitbucket.org/nxg/pqrst
 */

/// @cond NO_DOCUMENT
#ifndef PQRST_DEBUG
#define PQRST_DEBUG 0
#endif
/// @endcond

#include "pqrst.h"

/** @file */ // for Doxygen

/// @cond NO_DOCUMENT

#if PQRST_TEST_MODE

// Within the unit-test harness.
// Assertions are available only in this case.
#include <stdio.h>
#include <assert.h>

#else
// Within the Arduino IDE (probably).

// This module is designed to be used on an Arduino.
// Thus we can't use assertions or exceptions in the normal case.
#undef assert
#define assert(x)
#endif

#ifdef ARDUINO
#include <Arduino.h>            // for Serial
#else
#include <stdio.h>              // being used in a test mode
#endif

// Define flags within the Task::flags field.  This is of type uint16_t.  See notes in pqrst.h.
#if PQRST_TASK_PRIORITIES
#define MASK_PRIORITY (1<<15)
#endif
// If this bit is set, then a task is rescheduled based on the current
// time, rather than the due time.
#define MASK_SLIPPY_TASK (1<<14)
// the low 10 bits are used for the estimated task duration
#define MASK_DURATION 0x3ff
#define MAX_DURATION  0x3ff

#if PQRST_SELF_QUEUE
#define QUEUE (*queue_)
#else
#define QUEUE Queue
#endif

const PROGMEM char* const Task::pqrst_version = "pqrst 1.0 (35be58e470c3), 2018 August 18: "
#if PQRST_TASK_PRIORITIES
    "prio,"
#else
    "noprio,"
#endif
#if PQRST_TASK_IDENT
    "ident,"
#else
    "noident,"
#endif
#if PQRST_SELF_QUEUE
    "self";
#else
    "noself";
#endif

// half-way through the time range
#define HALF_TIME 0x80000000

// suppress unused-parameter warnings
// (this works for both clang and GCC, but GCC demands the __attribute__ come after the variable)
#define UNUSED(x) UNUSED_ ## x __attribute__((__unused__))

/// @endcond

/**
 * Creates a new TaskQueue.
 */
TaskQueue::TaskQueue(void)
    : head(nullptr),
      reftime(0),
      // this initialisation is redundant with initialise() below,
      // but is necessary since ModularTimeMinder(Task*) doesn't have a default constructor
#if PQRST_SELF_QUEUE
      modtime_minder(ModularTimeMinder(this))
#else
      modtime_minder(ModularTimeMinder())
#endif
{
    initialise();
}

/**
 * Initialise a TaskQueue.  This will only be necessary if we want to
 * start using a Queue again, after it has been cleared by
 * {@link remove_all_tasks}.
 */
void TaskQueue::initialise(void)
{
    head = nullptr;
    reftime = 0;

#if PQRST_SELF_QUEUE
    modtime_minder = ModularTimeMinder(this);
#else
    modtime_minder = ModularTimeMinder();
#endif
    modtime_minder.start(HALF_TIME);
}

/// @cond NO_DOCUMENT

#if PQRST_SELF_QUEUE
ModularTimeMinder::ModularTimeMinder(TaskQueue* tqp)
    : MaintenanceTask(tqp),
#else
ModularTimeMinder::ModularTimeMinder()
    : MaintenanceTask(),
#endif
      fix(0)
{
#if PQRST_TASK_IDENT
    set_task_ident("modtime_minder");
#endif
};

/**
 * Manage the modular-time bit.  This task runs twice per time cycle:
 * at time HALF_TIME and time 0.  It flips the 'fix' bit and
 * reschedules itself.
 *
 * @param the current time
 */
void ModularTimeMinder::run(ms_t t)
{
    ms_t next_time;

    if (t >= HALF_TIME) {
        fix = HALF_TIME;
        next_time = 0;
    } else {
        fix = 0;
        next_time = HALF_TIME;
    }

    run_at(next_time);

#if PQRST_DEBUG
    fprintf(stderr, "modtime_minder @ 0x%x: fix=0x%x, next @ 0x%x\n", t, fix, next_time);
#endif
}

/// @endcond

/// @cond NO_DOCUMENT
#define MODTIME(t) QUEUE.modtime(t)
/// @endcond

/**
 * Impose an ordering on tasks.
 * The ordering respects the relative priority of the tasks, and their
 * anticipated durations.  If the tasks are of equal priority, then
 * this task compares earlier iff its duration is shorter.
 *
 * @param other the task which this task is being compared to
 * @return true if this task should be ordered before the other task
 */
#if PQRST_TASK_PRIORITIES
bool Task::before(const Task* other) const
{
    bool result;

    // get_priority returns either 0 or 1
    unsigned char my_priority = get_priority();
    unsigned char other_priority = other->get_priority();

    if (my_priority == other_priority) {
        // both high (the most common case), or both low
        result = (ready_time_ == other->ready_time_)
            ? get_duration() < other->get_duration()
            : MODTIME(ready_time_) < MODTIME(other->ready_time_);
    } else if (my_priority) {
        // this is higher priority
        result = MODTIME(ready_time_) < MODTIME(other->ready_time_ + other->get_duration());
    } else {
        // other is higher priority
        result = MODTIME(ready_time_ + get_duration()) <= MODTIME(other->ready_time_);
    }

#if PQRST_DEBUG && !ARDUINO
    if (ready_time_ >= 0x10000) { // not HALF_TIME, but 'large'
        fprintf(stderr, "before: (0x%x+%u)%s %c (0x%x+%u)%s (modtimes=0x%x,0x%x)\n",
                ready_time_, get_duration(), get_priority()>0?"!":"",
                result ? '<' : '>',
                other->ready_time_, other->get_duration(), other->get_priority()>0?"!":"",
                MODTIME(ready_time_), MODTIME(other->ready_time_));
    } else {
        fprintf(stderr, "before: (" MS_FMT "+" MS_FMT ")%s %c (" MS_FMT "+" MS_FMT ")%s\n",
                ready_time_, get_duration(), get_priority()>0?"!":"",
                result ? '<' : '>',
                other->ready_time_, other->get_duration(), other->get_priority()>0?"!":"");
    }
#endif

    return result;

}
#else
bool Task::before(const Task* other) const
{
    return MODTIME(ready_time_) < MODTIME(other->ready_time_);

}
#endif /* PQRST_TASK_PRIORITIES */
#undef MODTIME

/**
 * Set a task to be ‘slippy’.  Normally, a task is rescheduled based
 * on the time at which it was due -- this means that a repeating task
 * will not slip if it is occasionally slightly delayed in starting.
 * If this is undesirable behaviour -- if we want a possibly expensive
 * task to be rescheduled based on the actual time -- then we can set
 * it as a ‘slippy’ task.
 *
 * If a non-slippy task takes a significant amount of time, or if it
 * is appreciably delayed, then it may end up putting itself back on
 * the queue, due in the past.  In this case it will be _immediately_
 * due.  This can result in a pile-up of tasks, possibly uselessly,
 * and possibly overwhelming the processor.
 *
 * @param is_slippy_p true if the task should be made slippy
 * @return the previous value of this parameter
 * @see get_task_slippy
 */
bool Task::set_task_slippy(bool is_slippy_p)
{
    bool old_value = flags & MASK_SLIPPY_TASK;

    if (is_slippy_p) {
        flags |= MASK_SLIPPY_TASK;
    } else {
        flags &= ~MASK_SLIPPY_TASK;
    }

    return old_value;
}

/**
 * Determine whether a task is ‘slippy’.
 * @return true if the task is slippy
 * @see set_task_slippy
 */
bool Task::get_task_slippy(void) const
{
    return flags & MASK_SLIPPY_TASK;
}

/** Return true if the task queue is empty. */
bool TaskQueue::is_empty(void) const
{
    // The queue is empty if there is at least one non-maintenance
    // task in it.  This will almost always be true, and so the loop
    // below will almost always terminate after the first iteration.
    for (Task* p=head; p!=nullptr; p=p->next) {
        if (p->user_task_p())
            return false;
    }

    return true;
}

/**
 * Return the number of user tasks currently in the queue.
 * It is better to call {@link is_empty} to test whether the queue is
 * empty, than test whether `size() == 0`.
 *
 * @param inc_maintenance include maintenance tasks in the count
 * @return the queue length
 */
unsigned int TaskQueue::size(bool inc_maintenance) const
{
    unsigned int n = 0;

    for (const Task* p=head; p!=nullptr; p=p->next) {
        if (inc_maintenance || p->user_task_p()) {
            n++;
        }
    }

    return n;
}


/**
 * Push the given task onto the queue, due at time t.
 *
 * If the task is already present on the queue, then the earlier
 * scheduling is cancelled.  A task will only be scheduled on a
 * queue once (it can, and often will, reschedule itself, but at any
 * time it will be present on the queue only once).
 *
 * If there is more than one queue in the code, a Task _must not_ be
 * scheduled in more than one of them.
 *
 * Do not use `push_at(Task*, 0)` as a way of scheduling a task for
 * ‘immediately’.  That is not guaranteed to work for all values of
 * the reference time.  If you wish this, instead use
 * {@link push_after} with a zero offset, or {@link push_immediately}.
 *
 * We can be deemed to take ownership of the new object.
 *
 * It may not need saying, but scheduling tasks within an interrupt handler
 * is probably an _extremely_ bad idea.
 *
 * @param T the task to be queued, which must be non-null
 * @param t the due time
 * @see Task::run_at
 */
void TaskQueue::push_at(Task* T, ms_t t)
{
    // This mustn't be called within an interrupt handler.  If we
    // decided that was a good idea, then we'd have to turn interrupts
    // off for the duration of this method.  That might be Bold.

    // We depend on T being non-null.  If we weren't on an Arduino
    // we'd check this at this point, and throw an exception if it was null.
    // Instead, simply make this method a no-op in this case.
    if (T == nullptr) return;

    // It should be the case that is_enqueued_p corresponds accurately
    // to the task being in the queue or not.  It's possible, in
    // principle, for a task to be marked as enqueued without actually
    // being so (for example, remove_all_tasks doesn't dequeue
    // objects, it simply forgets about them).  It should not be
    // possible for a task to be in the queue but not marked as such.
    // The result of the former case, however, would be that we would
    // re-scan the queue to remove the non-present task, uselessly but
    // harmlessly (TaskQueue::remove can cope with this case).  This
    // case is probably fairly unlikely to happen other than in a
    // test harness.

    T->ready_time_ = t;

#if PQRST_TEST_MODE
    unsigned int qlen = size(true);
#endif

    if (T->is_enqueued_p()) {
        T->cancel();
#if PQRST_TEST_MODE
        qlen--;
#endif
    }

    assert(! T->is_enqueued_p());

    if (head == nullptr || T->before(head)) {
        T->next = head;
        head = T;

    } else {
        bool inserted_p = false;

        for (Task* p=head; !inserted_p; p=p->next) {
            assert(p != nullptr);
            if (p->next == nullptr || T->before(p->next)) {
                T->next = p->next;
                p->next = T;
                inserted_p = true;
            }
        }
    }

#if PQRST_DEBUG
    fprintf(stderr, "%s push_at 0x%x\n",
#if PQRST_TASK_IDENT
            T->get_task_ident(),
#else
            "task",
#endif
            t);
#endif

    assert(size(true) == qlen+1);

    assert(queue_is_valid_p());
    assert(T->is_enqueued_p());
}

/**
 * Push the given task onto the queue, due some (positive) interval
 * after the current reference time.
 *
 * Note that the reference time is _not_ `millis()`, but instead the
 * last time at which there was a non-null pop of the queue, or a
 * signal sent to the queue.  This is because we assume that
 * the majority of the calls to this method are happening within a
 * `run()` method, in which case the relevant time is the time when the
 * task was due, not the possibly slightly later actual time.  If you
 * need to push after the actual current time, or wish to push a task
 * outside a `run()` method, then use `push_at(millis()+t)`.
 *
 * Do not use this method outside a {@link Task::run} method.
 *
 * @param T the task to be queued
 * @param interval the time after the reference time, in non-negative ms, at which the task is to be queued
 */
void TaskQueue::push_after(Task* T, ms_t interval)
{
    // If this is called with a negative literal argument, it is
    // coerced to a large positive unsigned integer, which is likely
    // to confuse things.
    push_at(T, reftime + interval);
}

/**
 * Push the given task on the queue, due immediately
 * (in particular, at the current reference time).
 * This is broadly equivalent to {@link push_after(Task*, ms_t)}, with an
 * `interval` argument of zero, **except that** the
 * task is forced to the head of the queue, _even if the current head
 * task is overdue_.  Thus this task will be due to run the next time
 * that {@link run_ready} is called; this is useful if some action is
 * deemed particularly urgent.
 *
 * If, as is more usually the case, you wish the task to run immediately without this preemption,
 * then use {@link push_after} with a zero interval argument.
 *
 * @param T the task to be pushed
 */
void TaskQueue::push_immediately(Task* T)
{
    // this is a degenerate case of push_at, qv
    if (T == nullptr) return;

    if (T->is_enqueued_p()) {
        T->cancel();
    }

    assert(! T->is_enqueued_p());

    T->ready_time_ = reftime;
    T->next = head;
    head = T;

    assert(queue_is_valid_p());
    assert(T->is_enqueued_p());
}

/**
 * Pop the head task from the queue if the given time has passed.
 * If the head task isn't due, this does nothing, and returns nullptr.
 * If the head task is due, this sets the ‘reference time’ to be the
 * time at which the head task was due, and returns that task.
 * We can regard ownership as being passed to the caller.
 *
 * If the task is ‘slippy’, then the reference time is ‘now’.
 *
 * @param t the current time
 * @return the head task, or nullptr if the head task isn't yet due.
 * @see Task::set_task_slippy
 */
Task* TaskQueue::pop(ms_t t)
{
    Task* rval = nullptr;

    if (head != nullptr && modtime(head->ready_time_) <= modtime(t)) {
        // there is a head task, and it's ready
        rval = head;
        head = head->next;
        assert(rval != nullptr);
        rval->next = PQRST_NEXT_UNQUEUED_; // mark as unqueued
        if (rval->get_task_slippy()) {
            reftime = t;
        } else {
            reftime = rval->ready_time_;
        }

#if PQRST_DEBUG
        if (t > 0x10000) {
            printf("t=0x%x  ready_time_=0x%x  reftime=0x%x%s\n",
                   t, rval->ready_time_, reftime, rval->get_task_slippy() ? " slippy" : "");
        } else {
            printf("t=%u  ready_time_=%u  reftime=%u%s\n",
                   t, rval->ready_time_, reftime, rval->get_task_slippy() ? " slippy" : "");
        }
#endif
    }

    assert(queue_is_valid_p());
    assert(rval == nullptr || ! rval->is_enqueued_p());

    return rval;
}


/**
 * Remove from the queue an element which is == to the given Task.
 * It is OK to 'remove' a task which is not in the queue:
 * in this case, nothing happens.
 *
 * @param T the task to be removed
 */
void TaskQueue::remove(Task* T)
{
    // we do not require that T is queued
    // (ie, it's OK to 'remove' an unqueued task).
    // We assume, below, that a task cannot be in the queue more than once.

    if (T == nullptr) {
        // bad call, but make it a no-op
        return;

    } else if (head == nullptr) {
        assert(! T->is_enqueued_p());

    } else if (! T->is_enqueued_p()) {
        // nothing to do

    } else if (head == T) {
        head = head->next;
        T->next = PQRST_NEXT_UNQUEUED_;

    } else {
        for (Task* p = head; p != nullptr; p = p->next) {
            if (p->next != nullptr && p->next == T) {
                p->next = T->next;
                T->next = PQRST_NEXT_UNQUEUED_;
                break;          // JUMP OUT of loop
            }
        }
    }

    assert(queue_is_valid_p());
    assert(! T->is_enqueued_p());
}

/**
 * If the head task in the queue is ready at the given time
 * (ie, its due time is at or before that time), then run it.
 * @param t the current time
 * @return true if a task was ready and run
 */
bool TaskQueue::run_ready(ms_t t)
{
    Task* T = pop(t);

    if (T) {
        T->run_task(t);
        return true;
    } else {
        return false;
    }
}

/**
 * Return the interval between the queue reference time and the due-time of the task at the head of the queue.
 * If the head task is overdue, this returns zero, rather than a negative value.
 * Returns 0xffffffff if the queue is empty.
 *
 * @return the interval
 */
ms_t TaskQueue::next_due_after(void) const
{
    Task* p;
    // skip maintenance tasks
    for (p=head; p!=nullptr; p=p->next) {
        if (p->user_task_p()) break;
    }

    if (p == nullptr) {
        return 0xffffffff;

    } else {
        ms_t diff;

        if (p->ready_time_ > reftime) {
            diff = p->ready_time_ - reftime;
        } else if (reftime >= HALF_TIME && p->ready_time_ < HALF_TIME) {
            // the ready time is after the next wraparound
            diff = (p->ready_time_ + HALF_TIME) - (reftime - HALF_TIME);
        } else {
            // the task is overdue
            diff = 0;
        }

        return diff;
    }
}

/**
 * Return the interval between the given time and the queue's
 * ‘reference time’.
 *
 * Note: the current implementation of this method assumes that `t` is
 * not earlier than the reference time, and it will not give
 * particularly useful answers for times which are earlier.  A future
 * implementation of this method may change the semantics in this case.
 *
 * @param t the time of interest
 * @return how late we are behind the queue reference time
 */
int TaskQueue::lateness(ms_t t) const
{
    // we don't have to do anything complicated around times
    // 0x80000000 or 0xffffffff -- the modular arithmetic gets it right.
    return t - reftime;
}

/**
 * Send the signal to one task.  We do this recursively, so that the
 * signal is sent first to the _last_ task in the queue, and then
 * works forwards.  This means that the set of tasks which will
 * receive the signal is determined before any signal is sent, which
 * in turn means that if any signalled tasks respond by _adding_ tasks
 * to the queue, then those added tasks will not receive the signal.
 */
void TaskQueue::signal_to_task1(Task* T, int signal)
{
    assert(T != nullptr);       // TaskQueue::signal should guarantee this
    if (T->next != nullptr) {
        signal_to_task1(T->next, signal);
    }
    T->signal(signal, reftime);
}

/**
 * Pass the given integer on to each task in the queue.  This calls
 * {@link Task::signal} for each such task.  There is no significance
 * to the number, as far as this class is concerned.  This updates the
 * queue's ‘reference time’ to the time passed in the argument.
 *
 * It is unspecified what order the tasks in the queue are signalled,
 * but if the signalled tasks add tasks to the queue, those tasks
 * will not receive the signal.
 *
 * @param signal the integer to be passed on.
 * @param t the current reference time
 */
void TaskQueue::signal(int signal, ms_t t)
{
    if (head != nullptr) {
        reftime = t;
        signal_to_task1(head, signal);
    }
}

/**
 * Traverse the queue.
 * Run through the queue, calling the provided function once for each entry.
 * This is primarily for debugging; there is no way to change the
 * queue during this traversal.
 *
 * The callback prototype is either `typedef void
 * (*traverse_queue_cb_t)(ms_t, const char* ident, Task*)` or
 * `typedef void (*traverse_queue_cb_t)(ms_t, Task*)` depending on
 * whether `PQRST_TASK_IDENT` is defined
 * (see the `pqrst.h` code for further details).
 *
 * @param callback the callback function
 * @param max the maximum number of tasks to traverse (default: unlimited)
 */
void TaskQueue::traverse_queue(traverse_queue_cb_t callback, int max)
{
    for (Task* T = head; T != nullptr && max != 0; T = T->next, max--) {
#if PQRST_TASK_IDENT
        (*callback)(T->ready_time_, T->ident, T);
#else
        (*callback)(T->ready_time_, T);
#endif
    }
}

// Define a dump_queue_callback function, one way or another
#if ARDUINO
#if PQRST_TASK_IDENT
static void dump_queue_callback(ms_t t, const char* id, const Task* UNUSED(T))
{
    Serial.print("  ");
    Serial.print(id);
    Serial.print(" @ ");
    Serial.println(t);
}
#else
static void dump_queue_callback(ms_t t, const Task* UNUSED(T))
{
    Serial.print("  task @ ");
    Serial.println(t);
}
#endif
#else
#if PQRST_TASK_IDENT
static void dump_queue_callback(ms_t t, const char* id, const Task* UNUSED(T))
{
    if (t >= 0x10000) {
        fprintf(stderr, "  %s @ 0x%x\n", id, t);
    } else {
        fprintf(stderr, "  %s @ " MS_FMT "\n", id, t);
    }
}
#else
static void dump_queue_callback(ms_t t, const Task* UNUSED(T))
{
    if (t >= 0x10000) {
        fprintf(stderr, "  task @ 0x%x\n", t);
    } else {
        fprintf(stderr, "  task @ " MS_FMT "\n", t);
    }
}
#endif
#endif


/**
 * Dump the queue.
 * The output is a sequence of times and pointers to functions.
 * For debugging only, since the `printf` will do nothing useful otherwise.
 * This method is therefore not available unless the variable `PQRST_TEST_MODE` is set.
 * The method {@link traverse_queue} is declared, though, so some debugging is
 * possible in that case.
 */
void TaskQueue::dump_queue_contents(void) // for debugging
{
    traverse_queue(dump_queue_callback);
}

/**
 * Discard all of the tasks from the queue, without running them.
 *
 * We don't do deallocation of any tasks, and we don't call
 * {@link Task::cancel} on them.  Any tasks that were on the queue may
 * end up in an inconsistent internal state.  Thus this method is
 * really only useful for the case where we wish to break out of the
 * main {@link is_empty} loop, such as in an error
 * handler (using `longjmp`, perhaps).
 *
 * After this operation, {@link is_empty} returns true.  If
 * you wish to start using the queue again, then you must call
 * {@link initialise} on the queue.
 */
void TaskQueue::remove_all_tasks(void)
{
    // It might seem tempting to scan through the queue before
    // discarding tasks, in order to mark tasks as unqueued.  We
    // mustn't do this, however, since is it possible for this
    // function to be called when the tasks have gone out of scope
    // (for example after a `longjmp`), in which case
    // the tasks in the queue will be invalid, may have been reused,
    // and so may have garbage in their 'next' pointers.

    head = nullptr;

    // Resetting the reftime isn't a documented behaviour of this
    // object, but it's useful in the unit tests for this method to do
    // a complete reset of the object.
    reftime = 0;
}

#if PQRST_TEST_MODE
/**
 * Test that the queue is valid.  That is:
 *
 *   * everything in the queue is {@link is_enqueued_p()} and
 *   * the queue is not circular.
 *
 * Call as `assert(queue_is_valid_p());`.  It is believed to be
 * impossible for this not to evaluate to true.
 *
 * Note: this is intended to work _only_ when doing unit-testing.
 * It is _not_ intended to be used in production on an Arduino, when
 * assertions are disabled in any case.  This _might_ work more
 * generally (ie, when used on a non-Arduino platform), but the
 * circularity test is simple-minded, and could produce false
 * failures if a valid queue is too long.
 *
 * @return true if valid
 */
bool TaskQueue::queue_is_valid_p() const
{
    // If we wanted or needed to make this more sophisticated, we
    // could run through the queue checking that we never see any task
    // twice: that would check for circularity _and_ the claim that no
    // task can be enqueued more than once.

    int maxcount = 32; // largest plausible queue size when testing
    Task* p = head;

    while (p != NULL) {
        if(! p->is_enqueued_p()) {
            return false;
        }
        if (--maxcount == 0) {
            // we've been going too long
            return false;
        }
        p = p->next;
    }
    return true;
}
#endif

#if PQRST_SELF_QUEUE
/**
 * Construct a task, associated with a queue.
 *
 * If `PQRST_SELF_QUEUE` is zero (the default), then this constructor
 * has no argument, and the Task is associated with the global queue.
 *
 * @param tqp the queue this task will run on
 */
Task::Task(TaskQueue* tqp)
    : ready_time_(0),
      next(PQRST_NEXT_UNQUEUED_),
#if PQRST_TASK_IDENT
      ident("<task>"),
#endif
#if PQRST_TASK_PRIORITIES
      flags(MASK_PRIORITY | 0x1), // priority; duration 1
#else
      flags(0),                 // no flags or duration defined in this case
#endif
      queue_(tqp)
#else
/**
 * Construct a task.
 *
 * Note: if the preprocessor constant `PQRST_SELF_QUEUE` is non-zero
 * (the default is zero),
 * then the constructor takes an argument `TaskQueue*`, pointing to
 * the queue which this task should be associated with.  The same must
 * be true of each of the subclasses of this class.
 */
Task::Task(void)
    : ready_time_(0),
      next(PQRST_NEXT_UNQUEUED_),
#if PQRST_TASK_IDENT
      ident("task"),
#endif
#if PQRST_TASK_PRIORITIES
      flags(MASK_PRIORITY | 0x1) // priority; duration 1
#else
      flags(0)            // no flags or duration defined in this case
#endif
#endif
{
    // empty
}

/**
 * Start the task.
 * The task will first become due the given delay after now.
 * If this is omitted, then the task will become immediately due.
 * See {@link TaskQueue::push_at} for the interpretation of the due time.
 *
 * The default implementation of this method consists of queuing the
 * task to run after the given `delay`; that is, it is broadly
 * equivalent to `run_after(delay)`, though possibly clearer in intent.
 * If, however, there is any work that the task should do on its first run, that
 * can be neatly achieved by overriding this task.  In such a case, it is
 * likely that the overriding method will want to schedule the task
 * with {@link TaskQueue::push_after}, or {@link Task::run_after}
 * or, most straightforwardly, the parent {@link Task::start} method.  Thus:
 *
 * <pre>
 * void MyClass::start(ms_t delay)
 * {
 *     // do one-time stuff...
 *
 *     Task::start(delay);
 * }
 * </pre>
 *
 * @param delay the interval after which this task should run
 */
void Task::start(ms_t delay)
{
    QUEUE.push_after(this, delay);

#if PQRST_DEBUG
#if ARDUINO
    Serial.print("Task::start: after ");
    Serial.println(delay);
#else
    fprintf(stderr, "Task::start: after " MS_FMT "\n", delay);
#endif

    QUEUE.dump_queue_contents();

#endif
}

/**
 * Start the task immediately.
 *
 * This is equivalent to {@link Task::start(ms_t)} with an argument of
 * 0.  This is the most common way that tasks are started.
 *
 * We declare this as a separate method `start(void)`, rather than
 * using a default for the argument, so that it's possible for
 * subclasses to override `start(void)` rather than obliging them to
 * override `start(ms_t)` even when they will always ignore the argument.
 */
void Task::start(void)
{
    Task::start(0);
}

/**
 * Push a new task onto the queue, due at a given time.
 * If the `new_task` is omitted, this reschedules the current task.
 * See {@link TaskQueue::push_at} for the interpretation of the due time.
 *
 * Do not schedule events in the past.  This will generally work much
 * as expected, but it is not guaranteed to work correctly near the
 * time the `ms_t` counter overflows.
 *
 * @param next_time the time at which the task is to be run
 * @param new_task the task to be queued (or the current task, if this is omitted or null)
 */
void Task::run_at(ms_t next_time, Task* new_task)
{
    QUEUE.push_at(new_task == nullptr ? this : new_task, next_time);
}

/**
 * Push a new Task onto the queue, due some (positive) interval
 * after the queue's current reference time.
 * Thus this is equivalent to {@link run_at(ms_t,Task*)} when given a time
 * `delay` ms after the current reference time (cf the {@link run}
 * method for discussion of the reference time).
 *
 * Do not use this method outside a {@link Task::run} method.
 *
 * @param delay the time after the reference time, at which the task is to be run
 * @param new_task the task to be queued (or the current task, if this is omitted or null)
 * @see run_at
 * @see TaskQueue::push_after
 */
void Task::run_after(ms_t delay, Task* new_task)
{
    QUEUE.push_after(new_task == nullptr ? this : new_task, delay);
}

/**
 * Push a task onto the queue, due immediately.
 * This is broadly equivalent to {@link run_after(ms_t,Task*)} with a
 * zero `delay` argument (so that it will be due at the queue's ‘reference time’),
 * but see {@link TaskQueue::push_immediately}.
 *
 * @param new_task the task to be queued
 * @see TaskQueue::push_immediately
 */
void Task::run_immediately(Task* new_task)
{
    QUEUE.push_immediately(new_task == nullptr ? this : new_task);
}

/**
 * Cancel this task.  The task will be removed from the queue.
 *
 * It's OK for a task to `cancel` itself, and this is a way for a
 * LoopTask to terminate the loop.
 *
 * You may `cancel` a task which is not scheduled – this is a no-op.
 *
 * @return the time at which the task would have been due, or zero if it was not queued
 */
ms_t Task::cancel(void)
{
#if PQRST_DEBUG && ARDUINO
    Serial.print("Task::cancel: ");
    Serial.println(ident);
#endif

    ms_t scheduled_time;
    if (is_enqueued_p()) {
        scheduled_time = ready_time_;
    } else {
        scheduled_time = 0;
    }

    QUEUE.remove(this);
    assert(!is_enqueued_p());

    return scheduled_time;
}

/**
 * Uncancel a task.  After a task has been removed from the queue
 * using {@link cancel}, it may be put back on the queue with this
 * method, and it will again be due at the time it was last scheduled.
 * If that time was in the past -- ie, the task is now overdue --
 * it will be due immediately.
 *
 * If the task is currently scheduled (ie, it has not been cancelled),
 * this method is a no-op.
 *
 * It is currently unspecified what the method returns if the task is
 * now overdue
 *
 * @return the new ready-time, if the task is ready in the future
 */
ms_t Task::uncancel(void)
{
#if 1
    QUEUE.push_at(this, ready_time_);
    return ready_time_;

#else
    // the following is probably a better interface, but I can't
    // straightforwardly work out how to test it (shortage of time,
    // rather than it being something deep).  QUEUE.lateness
    // will only work if a task has run, and a simple test_uncancel
    // test doesn't do that...
    ms_t rval = 0;

    if (! is_enqueued_p()) {
        // This test is somewhat redundant, since if the task is
        // currently scheduled, the following would remove it, and
        // then reschedule it at the same time.  However doing the
        // test probably saves a few cycles, and also makes the
        // meaning clearer.
        printf("uncancel %s: ready=%u  lateness=%u\n", ident, ready_time_, QUEUE.lateness(ready_time_));
        if (QUEUE.lateness(ready_time_) > 0) {
            QUEUE.push_after(this, 0);
            rval = 0;
        } else {
            QUEUE.push_at(this, ready_time_);
            rval = ready_time_;
        }
    }

    return rval;
#endif
}

/**
 * Receive a signal.  This is a *protected* method, which should be
 * overridden by an implementation if it wishes to take action on signals.
 *
 * The task queue can be given a signal to pass on to enqueued tasks.
 * The queue then runs through the currently waiting tasks and invokes
 * this method for each one.  The default method does nothing, so a
 * task should override this method only if it wishes to take action
 * in this case.
 *
 * @param signal the numeric signal to be passed on
 * @param t the effective time the signal was received
 *
 * @see TaskQueue::signal
 */
void Task::signal(int UNUSED(signal), ms_t UNUSED(t))
{
    return;
}

/**
 * Run this task; this is the method which is actually called when a
 * task is due to be run.
 * This calls the {@link run} method, possibly doing some
 * class-specific housekeeping as it does so.
 *
 * In some circumstances, it may be necessary for a subclass of
 * {@link Task} to override this.
 * For example, the {@link LoopTask::run_task} method decrements a
 * counter before calling its parent's `run_task()` method.  This
 * means that implementations of `run()` need not worry about running
 * superclass methods.
 */
void Task::run_task(ms_t t)
{
    run(t);
}

#if PQRST_TASK_PRIORITIES
/**
 * Set the priority of the task.  A task can (currently) only have one
 * of two priorities.  Any priority above 1 is taken to be 1 (which is
 * higher priority than 0).
 *
 * We wish to avoid the situation where a long running ‘background’ task
 * (a number of ms) is still running when a high-priority task is scheduled.
 * If, therefore, when we add a task to the queue, we would end up
 * with a high-priority task scheduled after a low-priority task with a
 * duration which extends past the high-priority start-time, then we
 * enqueue them so that the high-priority task is run (at its due time) before the
 * low-priority one.
 *
 * Two equal-priority tasks are scheduled with the shorter going
 * first, or the first to be added going first if their durations are equal.
 *
 * A low-priority, ‘long’, task should have a
 * duration declared, and be marked as a non-priority task, with
 * `set_priority(0)`; the duration need only be an estimate for these
 * scheduling purposes.
 *
 * The algorithm is relatively simple, but should suffice.  The goal
 * is to resolve occasional collisions between tasks, and the
 * algorithm doesn't attempt to resolve three-way collisions where,
 * for example, a high-priority task ends up enqueued behind a task
 * which might be in turn delayed by a previous long-running task.  At
 * that point, you have other problems.
 *
 * By default, tasks are created as priority/non-background tasks of duration 1.
 *
 * This method, and the associated priority-sensitive queue ordering,
 * is available only if `PQRST_TASK_PRIORITIES` is non-zero.
 *
 * @param task_priority the set priority, either zero or one
 * @see set_duration
 * @see get_priority
 */
void Task::set_priority(unsigned char task_priority)
{
    if (task_priority) {
        flags |= MASK_PRIORITY;
    } else {
        flags &= ~MASK_PRIORITY;
    }
}

/**
 * Return the priority of the task.
 *
 * This method is available only if `PQRST_TASK_PRIORITIES` is non-zero.
 *
 * @return 1 if the task is normal (high) priority, 0 otherwise
 * @see set_priority
 */
unsigned char Task::get_priority(void) const
{
    // for eg Task::before to work properly, this must return
    // specifically 0 or 1, and not merely zero/nonzero
    return flags & MASK_PRIORITY ? 1 : 0;
}

/**
 * Set the duration of the task.  This is only an estimate.
 * There is a maximum duration of (currently) 1023ms.
 *
 * This method is available only if `PQRST_TASK_PRIORITIES` is non-zero.
 *
 * @param duration the task's expected duration
 * @see set_priority
 * @see get_duration
 */
void Task::set_duration(ms_t duration)
{
    if (duration > MAX_DURATION)
        duration = MAX_DURATION;

    flags = (flags & ~MASK_DURATION) | duration;
}

/**
 * Return the task's expected duration.
 *
 * This method is available only if `PQRST_TASK_PRIORITIES` is non-zero.
 *
 * @return the duration in ms
 * @see set_duration
 */
ms_t Task::get_duration(void) const
{
    return flags & MASK_DURATION;
}
#endif /* PQRST_TASK_PRIORITIES */

#if PQRST_SELF_QUEUE
/**
 * Construct a looping task.  The task is rescheduled to run at the
 * given cadence.
 *
 * If `repeats` is positive, then the task is
 * rescheduled only the given number of times; if it is negative, then
 * it is rescheduled indefinitely; and if it is zero, it is never
 * rescheduled, and thus acts the same as a simple Task.
 *
 * If the `cadence` is zero, the task is never rescheduled, and thus
 * acts the same as a simple Task.
 *
 * The task is rescheduled the given interval after the _due_ time; if
 * there is any delay in running the task, then the gap between actual
 * run times may vary slightly from this on short timescales, but it shouldn't drift.
 *
 * The task is not automatically scheduled.  Once created, it must be
 * initially scheduled with {@link Task::start}, and it will repeat at the
 * given cadence thereafter.
 *
 * If `PQRST_SELF_QUEUE` is zero (the default), this
 * constructor omits the first `TaskQueue*` argument.
 *
 * @param tqp the queue this task is to be associated with
 * @param cadence the interval in ms between repeats
 * @param repeats the number of repeats; if negative, repeat indefinitely
 */
LoopTask::LoopTask(TaskQueue* tqp, ms_t cadence, int repeats)
    : Task(tqp),
      cadence(cadence),
      n(repeats)
#else
      // the following documentation is identical, except for the
      // missing tqp param
/**
 * Construct a looping task.  The task is rescheduled to run at the
 * given cadence.
 *
 * If `repeats` is positive, then the task is
 * rescheduled only the given number of times; if it is negative, then
 * it is rescheduled indefinitely; and if it is zero, it is never
 * rescheduled, and thus acts the same as a simple Task.
 *
 * If the `cadence` is zero, the task is never rescheduled, and thus
 * acts the same as a simple Task.
 *
 * The task is rescheduled the given interval after the _due_ time; if
 * there is any delay in running the task, then the gap between actual
 * run times may slightly vary from this.
 *
 * The task is not automatically scheduled.  Once created, it must be
 * initially scheduled with {@link Task::start}, and it will repeat at the
 * given cadence thereafter.
 *
 * Note: if the preprocessor constant `PQRST_SELF_QUEUE` is non-zero,
 * then the constructor takes an additional first argument
 * `TaskQueue*` (see {@link Task}).
 *
 * @param cadence the interval in ms between repeats
 * @param repeats the number of repeats; if negative, repeat indefinitely
 */
LoopTask::LoopTask(ms_t cadence, int repeats)
    : Task(),
      cadence(cadence),
      n(repeats)
#endif
{
    // empty
}

/**
 * Set the scheduling cadence.
 * Setting this to zero stops the loop from being re-scheduled
 * (see also {@link cancel}).
 * @param new_cadence the new cadence
 * @return the previous cadence
 */
ms_t LoopTask::set_cadence(ms_t new_cadence)
{
    ms_t old_cadence = cadence;
    cadence = new_cadence;
    return old_cadence;
}

/**
 * Set the repeat-count of the loop.  This resets the count; it does
 * not take any account of the number of repeats already executed.
 * The repeat count has the meanings described in the constructor.
 *
 * @param new_n the new repeat count
 * @return the previous repeat count
 * @see LoopTask::LoopTask
 */
int LoopTask::set_repeats(int new_n)
{
    int old_n = n;
    n = new_n;
    return old_n;
}

/**
 * Run the task, additionally rescheduling it.  This method is public
 * so that it can be overridden if necessary in some special
 * circumstance; but that will be unusual, and should probably be done
 * only after examination of the source code.
 *
 * That said, the following shows how one might dynamically adjust the
 * cadence of a loop task:
 * <pre>
 * void MySpecialLoopTask::run_task(ms_t t)
 * {
 *     if (adjust_cadence_this_time()) {
 *         ms_t orig_cadence = set_cadence(2);
 *         LoopTask::run_task(t);
 *         set_cadence(orig_cadence);
 *     } else {
 *         LoopTask::run_task(t);
 *     }
 * }
 * </pre>
 */
void LoopTask::run_task(ms_t t)
{
    /* Implementation note:  First reschedule the task if required,
     * then run it.   Doing it in this order means that the task can
     * cancel itself.
     */
    if (n > 0) {
        --n;
    }
    if (n != 0 && cadence > 0) {
        run_after(cadence, this);
    }

    Task::run_task(t);
}

/**
 * Construct a task which will run the supplied function at a later time.
 *
 * The function in question has a rather restricted form.  It must
 * match the type `typedef void(*run_task_t)(ms_t t)`, which means
 * that it cannot be an instance method; and it cannot refer to any
 * local automatic variables.  It can, however, be a simple function
 * or a static method, and it can refer freely to global variables.
 *
 * Note that the function can also usefully be a C++ lambda, as in:
 * <pre>
 * FunctionTask([](ms_t) { ...; }).run_after(10);
 * </pre>
 * This will run the contents of the function body 10ms after the
 * current time.
 *
 * This class extends the {@link LoopTask}, but unlike the constructor for
 * that task, the `cadence` is optional and defaults to zero, meaning
 * that by default the task will run only once.
 *
 * If `PQRST_SELF_QUEUE` is non-zero (the default is zero), this
 * constructor includes an initial `TaskQueue*` argument (see {@link Task}).
 *
 * @param f the function to be run
 * @param cadence the interval between runs, or zero to suppress any repetition (default: 0)
 * @param repeats the number of repeats before being cancelled (default: -1)
 */
#if PQRST_SELF_QUEUE
FunctionTask::FunctionTask(TaskQueue* tqp, run_task_t f, ms_t cadence, int repeats)
    : LoopTask(tqp, cadence, repeats),
      the_function(f)
      // * @param tqp the queue this task is associated with
#else
FunctionTask::FunctionTask(run_task_t f, ms_t cadence, int repeats)
    : LoopTask(cadence, repeats),
      the_function(f)
#endif
{
    // empty
}
// Note on closures:
//
// The following examples don't work, because these mutable lambdas
// don't match the required function prototype.  We need them to be
// declared mutable, so that the system can form a closure around them.
// The C++ library on the Arduino (which is where this pqrst library
// is targeted) doesn't have support for `std::function`, and nor does
// it have support for `std::unique_ptr`, which (I _think_) is
// necessary to implement even a cut-down `std::function` locally
// (see the discussion in
// [this blog post](https://shaharmike.com/cpp/naive-std-function/)
// about a 'naive' implementation of `std::function`).
//
// An apparent alternative would be to arrange the code such that
// run_at and run_after were const (because then we wouldn't need the
// 'mutable'), but trying to mark them as such -- because queueing a
// task does seem to be a logically const thing -- seems to require
// rather intricate changes to constness throughout the code, which
// I'm not sure ends up sensible.
//
// FunctionTask f1 = FunctionTask(TQPP [t1](ms_t t) mutable { t1.run_at(TIME_IMMEDIATELY); });
// FunctionTask f2 = FunctionTask(TQPP [t1](ms_t t) mutable { t1.run_at(8); });
// FunctionTask f3 = FunctionTask(TQPP [t2](ms_t t) mutable { t2.run_after(0); });



void FunctionTask::run(ms_t t)
{
    the_function(t);
}

#if !PQRST_SELF_QUEUE
// The global task queue
TaskQueue Queue;
#endif
