/*
 * Copyright (c) 2017, Norman Gray
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms,
 * with or without modification,
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * pqrst 1.0, revision 35be58e470c3, of 2018 August 18
 * See https://bitbucket.org/nxg/pqrst
 */

// Priority Queue for Running Simple Tasks
#ifndef PQRST_H_LOADED
#define PQRST_H_LOADED 1

/** @file */ // for Doxygen

#include <inttypes.h>

/*
 * You may wish to configure the following three defines,
 * by editing this file or by prepending settings for them.
 */

#ifndef PQRST_SELF_QUEUE
/**
 * If `PQRST_SELF_QUEUE` is set non-zero, then each Task must be
 * associated with a TaskQueue when it is created; if not, then the
 * library declares a global queue, called `Queue`, which all tasks
 * are scheduled on, and which must be defined by the calling program.
 * @hideinitializer
 */
#define PQRST_SELF_QUEUE 0
#endif

#ifndef PQRST_TASK_PRIORITIES
/**
 * If `PQRST_TASK_PRIORITIES` is set non-zero, then each
 * task has a priority (currently only 0 or 1) which allows it to
 * pre-empt a lower-priority task which might still be running when
 * the high-priority task is due (see the {@link Task::set_priority} method for
 * more discussion).  The default is for this to be zero.  If you do
 * need this functionality, then you can set `PQRST_TASK_PRIORITIES`
 * to one, at the cost of a few clock cycles.
 *
 * (The `priority` in this module's name refers to the time priorities
 * of the task queue, as distinct from these task priorities)
 * @hideinitializer
 */
#define PQRST_TASK_PRIORITIES 0
#endif

#ifndef PQRST_TASK_IDENT
/**
 * If `PQRST_TASK_IDENT` is set non-zero, then each Task object has an
 * identification string which can be set with {@link Task::set_task_ident} and
 * retrieved with {@link Task::get_task_ident}, and which will be reported at
 * various places if `DEBUG` is true.  See also
 * {@link TaskQueue::traverse_queue}, the callback for which changes its
 * signature in this case.
 * @hideinitializer
 */
#define PQRST_TASK_IDENT 0
#endif

/* No configuration below */

// forward class definition
class TaskQueue;

/**
 * The Arduino supports a time type of `unsigned long`, which is a
 * 4-byte unsigned integer on that platform.  We stick with that size here, for
 * convenience, and refer to it as `ms_t`.  Note that, although this
 * is referred to as ‘milliseconds’ here and elsewhere in this
 * documentation, this can in practice be any 32-bit time value: `millis()` or
 * `micros()` or any other jiffies that may be convenient, as long as
 * the times given to the various push/run methods, and as long as the
 * time given to {@link TaskQueue::run_ready} are in consistent units.
 */
typedef uint32_t ms_t;

/**
 * A printf format specifier which provides the correct spec for a
 * value of type ms_t
 */
#if !ARDUINO && !defined(MS_FMT)
// this is used only in non-Arduino debugging code
#define MS_FMT "%u"
#endif

/// @cond NO_DOCUMENT
#if ARDUINO
#include <avr/pgmspace.h>
#else
#define PROGMEM
#endif

/**
 * `TIME_NEVER` is a time which will never happen.
 *
 * Well, actually, To be precise, this _will_ happen at 0xffffffff
 * ms, or 49.7 days, at the time the Arduino clock wraps around.  The
 * PQRST code has no problem with the wrap-around, but (as things
 * stand now) a `TIME_NEVER` task would be run at this point.
 *
 * FIXME: handle never-happen events more consistently?  Simply delete this?
 */
#ifndef TIME_NEVER
#define TIME_NEVER ((ms_t)0xffffffff)
#endif

// Marker for a Task being queued/unqueued:
// Task.next == PQRST_NEXT_UNQUEUED_ if the Task is _out_ of the queue.
// See method is_enqueued_p().
#define PQRST_NEXT_UNQUEUED_ reinterpret_cast<Task*>(1)
/// @endcond

/**
 * A Task represents a block of code to be executed at some point in the future.
 *
 * An object which is to be queued in this way should extend the
 * `Task` class, and implement the {@link run} method.
 *
 * If `PQRST_SELF_QUEUE` is zero (the default), then there is a single
 * global queue, called `Queue`, on which tasks are queued by default.
 * If `PQRST_SELF_QUEUE` is non-zero, then a Task must be associated
 * with a TaskQueue when it is created, and each of the task
 * constructors takes an initial `TaskQueue*` pointer as a first
 * argument.  A Task can reschedule itself on
 * this queue, or schedule other tasks there.
 *
 * After the task is created, it must be started (that is, placed on
 * to the queue) with a call to {@link start}, or possibly {@link run_after}.
 */
class Task {
    friend class TaskQueue;     // TaskQueue needs to see ready_time_ and next

public:
    /**
     * Reports the PQRST version and compilation options.
     * This is in flash memory (Arduino `PROGMEM`), so should be printed appropriately.
     */
    static const PROGMEM char* const pqrst_version;

private:
    ms_t ready_time_;   ///< The time at which this task will be ready
    Task* next;         ///< The next task in the queue

#if PQRST_TASK_IDENT
    const char* ident;  ///< An identifier for this Task
#endif

    /**
     * Task flag bits.  At present, we dedicate 10 of these bits to
     * the duration (giving a maximum duration of 1023ms).  This
     * leaves 6 flag bits for use as flags, only two of which we use
     * at present.  For the actual usage of bits in this variable, see
     * pqrst.cpp.
     */
    uint16_t flags;

#if PQRST_SELF_QUEUE
    TaskQueue* queue_; ///< The queue this task is associated with
#endif

protected:
    virtual void signal(int signal, ms_t t);
    virtual void run_task(ms_t);

public:
#if PQRST_SELF_QUEUE
    Task(TaskQueue*);
#else
    Task();
#endif
    virtual ~Task() {};

    virtual void start(ms_t delay);
    virtual void start(void);

#if PQRST_SELF_QUEUE
    /**
     * Returns the queue this task is associated with.
     *
     * This is available only if `PQRST_SELF_QUEUE` is defined as 1.
     */
    TaskQueue* queue(void) const { return queue_; };
#endif

    /**
     * Runs one step of the task.
     *
     * The task has the option of re-queuing a task on the queue it
     * was retrieved from; it might re-queue itself
     * or another task, with {@link run_after(ms_t,Task*)}.
     *
     * The run method is given the current time, as passed to the
     * {@link TaskQueue::run_ready} method.  This may be different
     * from (later than) the time at which the task was due, if there
     * was any delay in running the queue.  The time at which the task
     * was due becomes the queue's ‘reference time’ (see {@link TaskQueue}).
     *
     * The task should regard the ‘current time’ as the time passed to
     * it in the argument, and should typically _not_ examine the time
     * by calls to `millis()` or `micros()`.
     *
     * @param t the current time for this task
     * @see set_task_slippy
     */
    virtual void run(ms_t t) = 0;

    ms_t cancel(void);
    ms_t uncancel(void);
    void run_at(ms_t t, Task* T=nullptr);
    void run_after(ms_t t, Task* T=nullptr);
    void run_immediately(Task* T=nullptr);

    bool before(const Task*) const;

    bool set_task_slippy(bool);
    bool get_task_slippy(void) const;

    /**
     * Return true if the task is currently scheduled on a queue.
     * @return true if enqueued
     */
    bool is_enqueued_p(void) const { return next != PQRST_NEXT_UNQUEUED_; };

    /**
     * Return the time when the tast is next ready.  This should be
     * used only for debugging, and the like, and should never be used
     * for `run_at` calculations.
     *
     * @return the next due time
     */
    ms_t ready_time(void) const { return ready_time_; };

#if PQRST_TASK_PRIORITIES
    void set_priority(unsigned char);
    unsigned char get_priority(void) const;
    void set_duration(ms_t);
    ms_t get_duration(void) const;
#endif

    /**
     * True if this is a ‘user task’, as opposed to a maintenance task.
     * This is not likely to be generally useful to user code, other than
     * perhaps in a queue-traversal callback.
     *
     * @return true if this is a user task
     */
    virtual bool user_task_p(void) const { return true; };

#if PQRST_TASK_IDENT
    /**
     * Set the identification string for a task.
     *
     * This is available only if `PQRST_TASK_IDENT` is defined.
     *
     * @param ident_string identification
     * @see get_task_ident
     */
    void set_task_ident(const char* ident_string) { ident = ident_string; };
    /**
     * Retrieve the identification string.
     *
     * This is available only if `PQRST_TASK_IDENT` is defined.
     *
     * @return the string
     * @see set_task_ident
     */
    const char* get_task_ident(void) const { return ident; };
#endif
};

/**
 * There are a few ‘maintenance tasks’ on the queue, placed there by
 * the queueing system.  User code cannot generate these tasks, but it
 * may occasionally be useful to detect whether a task is thus classed
 * – perhaps within a {@link TaskQueue::traverse_queue} callback – and
 * the method {@link Task::user_task_p} can be used for that.
 */
class MaintenanceTask : public Task {
    friend class ModularTimeMinder;

private:
#if PQRST_SELF_QUEUE
    MaintenanceTask(TaskQueue* tqp) : Task(tqp) {};
#else
    MaintenanceTask() : Task() {};
#endif

public:
    /**
     * Indicates whether this is a user task or a maintenance task.
     * All instances of this class return false.
     * @return false
     */
    bool user_task_p(void) const { return false; };
};

/**
 * A Task which will automatically reschedule itself after it is run.
 */
class LoopTask : public Task {
private:
    /**
     * The interval between loop runs, in ms.
     */
    ms_t cadence;
    /**
     * The number of repeats still to do.
     * If this is negative, repeat indefinitely.
     */
    int n;

protected:
    void run_task(ms_t);

public:
#if PQRST_SELF_QUEUE
    LoopTask(TaskQueue*, ms_t cadence, int repeats=-1);
#else
    LoopTask(ms_t cadence, int repeats=-1);
#endif

    ms_t set_cadence(ms_t);
    int set_repeats(int n);

    /**
     * Return the interval between loop runs.
     * @return an interval in ms
     * @see set_cadence
     */
    ms_t get_cadence(void) const { return cadence; };
    /**
     * Returns the number of repeats still to do.
     * If this is negative, the task is to be repeated indefinitely.
     * @return repeat count
     * @see set_repeats
     */
    int get_repeats(void) const { return n; };

    /**
     * Run a single iteration of the loop.  The run may schedule other
     * tasks, but will typically not reschedule itself, since this is
     * handled automatically.  If, however, the task has a `cadence`
     * or `repeats` of zero, so that it is not automatically
     * rescheduled, then this method may need or want to do so as appropriate.
     */
    virtual void run(ms_t) = 0;
};

/**
 * The type of the function which can be supplied to a {@link FunctionTask}.
 * When the task is ready, the function is called with the current
 * reference time as argument – the function should typically _not_
 * call `millis()` or `micros()`.
 *
 * @param the reference time
 */
typedef void(*run_task_t)(ms_t t);

/**
 * A task which runs a supplied function at a given time.
 *
 * The supplied function must match the type {@link run_task_t}, and
 * can be a predefined function or a lambda.  Note, however, that in
 * this version of the code the lambda _cannot_ declare any objects in
 * its closure.
 */
class FunctionTask : public LoopTask {
private:
    run_task_t the_function;

public:
#if PQRST_SELF_QUEUE
    FunctionTask(TaskQueue*, run_task_t, ms_t cadence=0, int repeats=-1);
#else
    FunctionTask(run_task_t, ms_t cadence=0, int repeats=-1);
#endif
    void run(ms_t);
};

/**
 * The type of the callback for {@link TaskQueue::traverse_queue}.  If
 * `PQRST_TASK_IDENT` is false, then this has two arguments, giving
 * the due time and a pointer to the Task; if the define is true, then
 * it has three arguments, where the second is the task identifier.
 */
#if PQRST_TASK_IDENT
/**
 * The type of the callback passed to the {@link TaskQueue::traverse_queue} method.
 * This is called once for each task in the queue.
 *
 * @param due the current time
 * @param ident the identification string of the task
 * @param T the task
 */
typedef void (*traverse_queue_cb_t)(ms_t due, const char* ident, const Task* T);
#else
/**
 * The type of the callback passed to the {@link TaskQueue::traverse_queue} method.
 * This is called once for each task in the queue.
 *
 * @param due the current time
 * @param T the task
 */
typedef void (*traverse_queue_cb_t)(ms_t due, const Task* T);
#endif

/// @cond NO_DOCUMENT

// The ms_t type (32-bit unsigned integer) will overflow at time
// 0xffffffff (=4294967295), half of which is 0x7fffffff
// (=2147483647).  Call the times below 0x80000000 'lower-time' and
// that time and above 'upper-time'.  We want time to be regarded as
// a ring, so that when the current time is a lower time, then all
// upper-times are in the future, and when the current time is an
// upper time, then all lower-times are in the future.
//
// Implement this in time comparisons (ie, within Task::before and
// TaskQueue::pop) by XORing times with ModularTimeMinder.fix, which
// we arrange to be zero during lower-time and 0x80000000 during
// upper-time.  In the former case, all lower times compare before all
// upper times; and in the latter case, all upper times compare before
// all lower times.

/**
 * A maintenance task, for ensuring that time is compared in a modular
 * fashion.  This is a maintenance task, and has no public constructor.
 */
class ModularTimeMinder : public MaintenanceTask {
    friend class TaskQueue;

private:
    ms_t fix;

    // private constructor -- used only by friend class TaskQueue
#if PQRST_SELF_QUEUE
    ModularTimeMinder(TaskQueue*);
#else
    ModularTimeMinder();
#endif
    void run(ms_t t);
};

/// @endcond


/**
 * A priority queue of Tasks.  The queue is maintained in time order;
 * if `PQRST_TASK_PRIORITIES` is defined as 1, it additionally takes
 * account of a high/low priority task flag.
 *
 * Tasks are added to the queue with a ‘ready time’ which is either
 * absolute, or relative to the queue's current ‘reference time’.  This
 * reference time is notionally ‘now’, but it is set to be a task's
 * ‘ready time’ when a task is run.  It may therefore be slightly
 * different from the ‘current time’ value passed to a {@link Task::run} method.  This
 * also means that two tasks which have the same ready
 * time will have the same reference time when run.  It is this
 * reference time which is used when calculating offsets using
 * {@link TaskQueue::push_after} and {@link Task::run_after}.
 *
 * It may not need saying, but scheduling tasks within an interrupt handler
 * is probably an _extremely_ bad idea (at least in this version of the code).
 *
 * **Note**: the reference time, as a reference for rescheduling, is
 * reliable only within the {@link Task::run} method, and so the
 * {@link push_after} and {@link Task::run_after} methods should only
 * be called within an implementation of that method.
 */
class TaskQueue {
private:
    Task* head;                 ///< Pointer to the head of the queue, and thus the earliest-due task
    ms_t reftime;               ///< The time of the last non-null pop()

    Task* pop(ms_t);
    void signal_to_task1(Task*, int);

    ModularTimeMinder modtime_minder;

#if PQRST_TEST_MODE
    // The following method should only be used in the case
    // PQRST_TEST_MODE.  Making sure it's undefined, unless we're in an
    // explicit test mode, means that it's almost impossible to have
    // assertions inadvertently enabled in a firmware (as opposed to
    // test) build (this is a good thing).
    bool queue_is_valid_p(void) const;
#endif


public:
    TaskQueue(void);
    bool is_empty(void) const;
    unsigned int size(bool inc_maintenance=false) const;
    void traverse_queue(traverse_queue_cb_t, int max=-1);

    void dump_queue_contents(void);       // should only be used for debugging
    void initialise(void);                // ditto

    bool run_ready(ms_t);
    ms_t next_due_after(void) const;
    void remove(Task*);
    void remove_all_tasks(void);
    void signal(int, ms_t);

    void push_at(Task*, ms_t);
    void push_after(Task*, ms_t);
    void push_immediately(Task*);

    int lateness(ms_t) const;

private:
    /**
     * Adjust the given time, for comparisons.
     * To cope with the timescale rolling over, we must adjust times
     * before comparing them (exclusively within
     * {@link Task::before}).  This method performs the appropriate adjustment.
     *
     * @param t task due time
     * @return adjusted time
     */
    ms_t modtime(ms_t t) const { return t ^ modtime_minder.fix; };

    // Allow the Task::before function to use this method, and nothing else
    friend bool Task::before(const Task* other) const;
};

#if !PQRST_SELF_QUEUE
/**
 * The global TaskQueue.  If `PQRST_SELF_QUEUE` is not defined,
 * then this is declared within the pqrst library,
 * and is globally visible as an `extern`.
 */
extern TaskQueue Queue;
#endif

#undef TQP

#endif /* PQRST_H_LOADED */
